#-----------------------------------------------------------------------
# This file is part of Nazca.
#
# Nazca is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# Nazca is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Nazca.  If not, see <http://www.gnu.org/licenses/>.

# @author: Ronald Broeke and Xaveer Leijtens (c) 2016-2017
# @email: ronald.broeke@brightphotonics.eu
#-----------------------------------------------------------------------

"""
This module defines mask elements such as straight and bent waveguides
via template functions.

Functions in mask_elements work with closures on nested functions. They
are templates functions and not specifically intended for use
in mask design directly by the designer. Instead use the interconnect
module and/or the predefined interconnects in
design kits.


However, a number of waveguide functions have been predefined in this
module for quick access and skipping Interconnect objects:

* strt      = Tp_straight()
* bend      = Tp_arc()
* ptaper    = Tp_ptaper()
* taper     = Tp_taper()
* cobra     = Tp_cobra()
* sinebend  = Tp_sinecurve()
"""

from math import sin, cos, acos, sqrt
from itertools import count
import warnings

import numpy as np

from .netlist import Cell
from . import cfg
from nazca.logging import logger
from .gds_base import gds_db_unit as gridsize
import nazca as nd
from nazca.util import polyline_length
from .generic_bend import curve2polyline, gb_point, gb_coefficients,\
    sinebend_point


min_length = 1e-12  # minimum straight waveguide length to draw


def layeriter(xs=None, layer=None):
    """Generator yielding all layers in a xsection.

    Args:
        xs (str): xsection name
        layer (int): layer number

    Yields:
        layer, growx, growy, accuracy: iterate over all layers in <xs> and <layer>
    """
    if xs is None and layer is None:
        xs = cfg.default_xs_name
        if xs not in cfg.XSdict.keys():
            _handle_missing_xs(xs)

    if layer is not None:
        grow = ((0.5, 0), (-0.5, 0), 0, 0) # a layer not as part of a xs has no grow
        layer_name = nd.get_layer(layer) #cfg.layer_names[layer][0:2]
        lineitem = cfg.layer_table.loc[layer_name]
        try:
            accuracy = lineitem['accuracy']
        except Exception as E:
            print("Error: accuracy not defined for layer {}.".format(layer))
            print(E)
            accuracy = 0.001
        yield layer_name, grow, accuracy, False

    if xs is not None:
        try:
            ML = cfg.XSdict[xs].mask_layers
        except:
            layer = _handle_missing_xs(xs)
            ML = cfg.XSdict[xs].mask_layers

        if ML.empty:
            warnings.warn("xsection '{0}' has no layers. "\
                "Continuing by adding fallback layer '{1}' to '{0}'.\n"\
                "Recommended solutions: Use a different xsection "\
                "or add layers to this xsection:\n"\
                "add_layer2xsection('{0}', layer=<num>).".\
                format(xs, cfg.default_layers['dump']), stacklevel=0)
            yield cfg.default_layers['dump'], ((0.5, 0), (-0.5, 0), 0, 0), 0.1
        else:
            for layer_name, A in ML.iterrows():
                grow = (
                    (A.leftedgefactor, A.leftedgeoffset),
                    (A.rightedgefactor, A.rightedgeoffset),
                    A.growy1,
                    A.growy2
                )
                yield layer_name, grow, A.accuracy, A.polyline


def _handle_missing_xs(xs):
    """Handle missing xsections or xsection missing mask_layers.

    If xsection <xs> is not defined, create a xsection with layer.
    A warning is issued, expect if the Nazca default xsection 'nazca' is created.

    If the xsection exists but has not mask_layer attribute this function
    will add a default layer to the xsection.
    The xsextions 'dump' and 'error' are special cases.

    Returns:
        layer: layer if any has been created in this method.
    """
    layer = None
    if xs not in cfg.XSdict.keys():
        if xs != cfg.default_xs_name and xs != cfg.default_xserror_name:
            logger.warning("No xsection named '{0}'. "\
                "Correct the name or add the xsection with:\n"\
                "add_xsection('{0}') to get rid of this warning. "\
                "Already available xsections are {1}.\n".\
                format(xs, list(cfg.XSdict.keys())))
        nd.add_xsection(xs)

    add_layer = False
    if not hasattr(nd.get_xsection(xs), 'mask_layers'):
        add_layer = True
    elif nd.get_xsection(xs).mask_layers is None:
        add_layer = True
    elif nd.get_xsection(xs).mask_layers.empty:
        add_layer = True
    if add_layer:
        if xs == cfg.default_xserror_name:
            layer = 'error'
        else:
            layer = 'dump'
        layer = nd.get_layer(layer)
        nd.add_layer2xsection(xs, layer=layer)
    return layer


#==============================================================================
# Waveguide element definitions
#==============================================================================
cnt = 0 #ordinal counter for unique naming
def Tp_straight(length=10, width=1.0, xs=None, layer=None, edge1=None,
        edge2=None, name=None):
    """Template for creating parametrized straight waveguide function.

    Args:
        length (float): length of waveguide
        width (float): width of waveguide
        xs (str): xsection of taper
        layer (int | str): layer number or layername
        edge1 (function): optional function F(t) describing edge1 of the waveguide
        edge2 (function): optional function G(t) describing edge2 of the waveguide

    Returns:
        function: Function returning a Cell object with a straight guide
    """

    def cell(length=length, width=width, xs=xs, layer=layer, edge1=edge1,
            edge2=edge2, name=name):
        """Create a straight waveguide element.

        Args:
            length (float): length of waveguide
            width (float): width of waveguide
            xs (str): xsection of waveguide
            layer (int | str): layer number or layer name
            edge1 (function): optional function F(t) describing edge1 of the waveguide
            edge2 (function): optional function G(t) describing edge2 of the waveguide

        Returns:
            Cell: straight element
        """
        if name is None:
            name = 'straight'
        # assert width is not None
        if width is None:
            width = 0.0
        with Cell(name=name, cnt=True) as C:
            C.instantiate = cfg.instantiate_mask_element
            p1 = nd.Pin(name='a0', type=0, width=width, xs=xs, show=True).put(0, 0, 180)
            p2 = nd.Pin(name='b0', type=1, width=width, xs=xs, show=True).put(length, 0, 0)
            nd.connect_path(p1, p2, length)
            C.length_geo = length # TODO: used by old trace method: to be removed

            for lay, grow, acc, polyline in layeriter(xs, layer):
                (a1, b1), (a2, b2), c1, c2 = grow
                if not polyline:
                    if edge1 is None:
                        outline = [
                            (0-c1, width*a1+b1), (length+c2, width*a1+b1),
                            (length+c2, width*a2+b2), (0-c1, width*a2+b2)
                        ]
                    else:
                        if edge2 is None:
                            edge2 = edge1
                        Fp1 = []
                        Fp2 = []
                        for t in np.linspace(0, 1, 20):
                            Fp1.append((length*t, edge1(t)*a1+b1))
                            Fp2.append((length*t, -edge2(t)*a2+b2))
                        outline = Fp1 + list(reversed(Fp2))
                    if abs(length) > min_length:
                        nd.Polygon(layer=lay, points=outline).put(0)
                else:
                    if abs(length) > min_length:
                        wpoly = width*(a1-a2)+(b1-b2)
                        centre = 0.5*(width*(a1+a2)+(b1+b2))
                        centreline = [(0-c1, centre), (length+c2, centre)]
                        nd.Polyline(layer=lay, points=centreline, width=wpoly).put(0)

        C.properties['parameters'] = {
              'length': {'value': length , 'unit': 'um', 'type': 'float'},
              'width':  {'value': width, 'unit': 'um', 'type': 'float'},
              'xs':     {'value': xs, 'unit': None, 'type': 'str'},
              'call':   {'value': 'strt', 'unit': None, 'type': 'str'},
         }
        return C
    return cell


def Tp_taper(length=100, width1=2, width2=3, xs=None, layer=None, name=None):
    """Template for creating a parametrized linear taper function.

    Args:
        length (float): length of the taper
        width1 (float): width at start
        width2 (float): width at end
        xs (str): xsection of taper
        layer (int | str): layer number or layer name

    Returns:
        function: Function returning a Cell object with a linear taper
    """

    def cell(length=length, width1=width1, width2=width2, xs=xs, layer=layer,
            name=name):
        """Create a taper element.

        Args:
            length (float): length of taper
            width1 (float): start width of taper
            width2 (float): end width of taper
            xs (str): xsection of taper
            layer (int | str): layer number or layer name

        Returns:
            Cell: taper element
        """
        if name is None:
            name = 'taper'
        with Cell(name=name, cnt=True) as C:
            C.instantiate = cfg.instantiate_mask_element
            if abs(length) < gridsize / 2:
                return C # Empty taper
            if width1 > width2:
                swap = True
                pin = ['b0', 'a0']
                width1, width2 = width2, width1
            else:
                swap = False
                pin = ['a0', 'b0']

            p1 = nd.Pin(name=pin[0], type=0, width=width1, xs=xs, show=True).put(0, 0, 180)
            p2 = nd.Pin(name=pin[1], type=1, width=width2, xs=xs, show=True).put(length, 0, 0)
            nd.connect_path(p1, p2, length)
            C.length_geo = length

            for lay, grow, acc, polyline in layeriter(xs, layer):
                (a1, b1), (a2, b2), c1, c2 = grow
                if swap:
                    (a1, b1), (a2, b2) = (-a1, -b1), (-a2, -b2)
                # Set widths for this layer
                twidth1a = a1*width1 + b1
                twidth1b = a2*width1 + b2
                twidth2a = a1*width2 + b1
                twidth2b = a2*width2 + b2
                if not polyline:
                    outline = [
                        (0-c1, twidth1a), (length+c2, twidth2a),
                        (length+c2, twidth2b), (0-c1, twidth1b)
                    ]
                    nd.Polygon(layer=lay, points=outline).put()
                else:
                    if abs(length) > min_length:
                        wpoly = twidth1a - twidth1b
                        centre = 0.5*(twidth1a + twidth1b)
                        centreline = [(0-c1, centre), (length+c2, centre)]
                        nd.Polyline(layer=lay, points=centreline, width=wpoly).put(0)

        C.properties['parameters'] = {
            'length': {'value': length, 'unit': 'um', 'type': 'float'},
            'width1': {'value': width1, 'unit': 'um', 'type': 'float'},
            'width2': {'value': width2, 'unit': 'um', 'type': 'float'},
            'xs':     {'value': xs, 'unit': None, 'type': 'str'},
            'call':   {'value': 'taper', 'unit': None, 'type': 'str'},
        }
        return C
    return cell


def Tp_ptaper(length=100, width1=1.0, width2=3.0, xs=None, layer=None,
        name=None, sigtype=None):
    """Template for creating a parametrized parabolic taper  function.

    Args:
        length (float): length of the taper
        width1 (float): width at start
        width2 (float): width at end
        xs (str): xsection of taper
        layer (int | str): layer number or layer name

    Returns:
        function: Function returning a Cell object with a ptaper
    """

    def cell(length=length, width1=width1, width2=width2, xs=xs, layer=layer,
            name=name):
        """Create a parabolic taper element.

        Args:
            length (float): length of taper
            width1 (float): start width of taper
            width2 (float): end width of taper
            xs (str): xsection of taper
            layer (int | str): layer number or layer name

        Returns:
            Cell: parabolic taper element
        """
        # TODO: add growy
        if name is None:
            name = 'ptaper'
        with Cell(name=name, cnt=True) as C:
            C.instantiate = cfg.instantiate_mask_element
            if abs(length) < gridsize / 2:
                return C # Empty taper
            if width1 > width2:
                swap = True
                pin = ['b0', 'a0']
                width1, width2 = width2, width1
            else:
                swap = False
                pin = ['a0', 'b0']

            p1 = nd.Pin(name=pin[0], type=0, width=width1, xs=xs, show=True).put(0, 0, 180)
            p2 = nd.Pin(name=pin[1], type=1, width=width2, xs=xs, show=True).put(length, 0, 0)
            nd.connect_path(p1, p2, length)
            C.length_geo = length

            for lay, grow, acc, polyline in layeriter(xs, layer):
                (a1, b1), (a2, b2), c1, c2 = grow
                if swap:
                    (a1, b1), (a2, b2) = (-a1, -b1), (-a2, -b2)
                if not polyline:
                    # Set widths for this layer
                    twidth1a = a1*width1 + b1
                    twidth1b = a2*width1 + b2
                    twidth2a = a1*width2 + b1
                    twidth2b = a2*width2 + b2
                    if abs(0.5*width1 - 0.5*width2) / (4*length) < gridsize:
                        # No shaping required due to small delta width:
                        points = [
                            (0, twidth1a), (length, twidth2a),
                            (length, twidth2b), (0, twidth1b)
                        ]
                        nd.Polygon(layer=lay, points=points).put(0)
                        continue
                    # y = a * x**2
                    a = 4 * length / (width2**2 - width1**2)
                    y2 = y0 = a * (0.5*width1)**2
                    w_tap1 = width1
                    ptop = [(0, twidth1a)]
                    pbot = [(0, twidth1b)]
                    while y2 - y0 < length:
                        w_tap2 = w_tap1 + 4*acc + 4*sqrt(acc*(w_tap1 + acc))
                        y2 = a * (0.5*w_tap2)**2
                        if y2 - y0 < length:
                            ptop.append((y2-y0, w_tap2*a1+b1))
                            pbot.append((y2-y0, w_tap2*a2+b2))
                            w_tap1 = w_tap2
                        else:
                            break
                    ptop.append((length, twidth2a))
                    pbot.append((length, twidth2b))
                    nd.Polygon(layer=lay, points=ptop + list(reversed(pbot))).put(0)
                else:
                    if abs(length) > min_length:
                        wpoly = width1*(a1-a2)+(b1-b2)
                        centre = 0.5*(width1*(a1+a2)+(b1+b2))
                        centreline = [(0-c1, centre), (length+c2, centre)]
                        nd.Polyline(layer=lay, points=centreline, width=wpoly).put(0)
        C.properties['parameters'] = {
            'length': {'value': length ,'unit': 'um', 'type': 'float'},
            'width1': {'value': width1, 'unit': 'um', 'type': 'float'},
            'width2': {'value': width2, 'unit': 'um', 'type': 'float'},
            'xs':     {'value': xs, 'unit': None, 'type': 'str'},
            'call':   {'value': 'ptaper', 'unit': None, 'type': 'str'},
        }
        return C
    return cell


def __get_offset(xs, width, radius):
    """Get the offset function."""
    offset = 0
    if xs is None or abs(radius) < 1e-3:
        return offset
    try:
        F = cfg.XSdict[xs].os
    except:
        _handle_missing_xs(xs)
        cfg.XSdict[xs].os = 0
    else:
        try:
            offset = float(F)
        except TypeError:
            offset = F(width=width, radius=radius)
    return offset


def Tp_arc(radius=10, width=1.0, angle=90, xs=None, layer=None, offset=None,
        name=None):
    """Template for creating a parametrized circular arc waveguide function.

    Args:
        radius (float): radius at the center line of the arc in um.
        width (float): width of the arc in um.
        angle (float): angle of arc in degree (default = 90).
        xs (str): xsection of taper
        layer (int | str): layer number or layer name
        offset (float | function): positive offset reduces radius.
            The offset can be a function F(width, radius) that returns a float

    Returns:
        function: Function returning a Cell object with an arc
    """
    # TODO: add growy
    def cell(radius=radius, width=width, angle=angle, xs=xs, layer=layer,
            offset=offset, name=name):
        """Create a circular arc element.

        A straight-bend offset is included when it has been defined in the
        xsection used.

        Args:
            radius (float): radius at the center line of the arc in um.
            width (float): width of the arc in um.
            angle (float): angle of arc in degree (default = 90).
            xs (str): xsection of taper
            layer (int | str): layer number or layer name
            offset (float | function): positive offset reduces radius.
                The offset can be a function F(width, radius) that returns a float

        Returns:
            Cell: circular arc element
        """
        if name is None:
            name = 'arc'
        ang = np.radians(angle)
        sign = np.sign(ang)
        radius = abs(radius)

        if offset is None:
            offset = __get_offset(xs, width, radius)

        Nmax = 298 # 2x298 + 4 = 600 (max in GDSII spec)
        with Cell(name=name, cnt=True) as C:
            C.instantiate = cfg.instantiate_mask_element
            p1 = nd.Pin(name='a0', type=0, width=width, xs=xs, show=True).put(0, 0, 180)
            p2 = nd.Pin(name='b0', type=1, width=width, xs=xs, show=True).\
                put(radius*sin(abs(ang)), sign*radius*(1-cos(ang)), angle)
            nd.connect_path(p1, p2, abs(radius*ang))
            C.length_geo = abs(radius*ang)

            if ang == 0:
                return C

            for lay, grow, acc, polyline in layeriter(xs, layer):
                (a1, b1), (a2, b2), c1, c2 = grow
                if angle >= 0:
                    (a1, b1), (a2, b2) = (-a1, -b1), (-a2, -b2)
                # Start and end sections are different from other sections
                if polyline:
                    wpoly = width*(a1-a2) + (b1-b2)
                    Ro = Ri = radius +0.5*(width*(a1+a2) + (b1+b2)) - offset
                else:
                    Ro = radius + width * a1 - offset + b1 # outer radius
                    Ri = radius + width * a2 - offset + b2 # inner radius
                    #TODO: add growy
                if Ri < 0:
                    logger.warning("Inner arc radius too small in cell '{}': inner_radius={:.3f}, xs='{}', width={:.3f}, offset={:.3f}. Setting it to 0.".\
                        format(cfg.cells[-1].cell_name, Ri, xs, width, offset))
                    Ri = 0
                da = 2 * acos(Ro / (Ro + acc)) # step angle for accuracy
                N = int((abs(ang) / da) + 1) # +1 for rounding
                da = ang / N # step angle for N
                u = np.linspace(da / 2, ang - da / 2, N)
                Roe = Ro * (da / 2) / sin(da / 2) # effective outer radius
                Rie = Ri * (da / 2) / sin(da / 2) # effective inner radius

                p1 = [(Roe * sin(abs(a)), sign * (radius - Roe * cos(a)))
                    for a in u]
                p2 = [(Rie * sin(np.abs(a)), sign * (radius - Rie * cos(a)))
                    for a in u]

                pstart = [(0, sign * (radius - Ri)), (0, sign * (radius - Ro))]
                pend = [
                    (Ro * sin(abs(ang)),
                    sign * (radius - Ro * cos(ang))),
                    (Ri * sin(abs(ang)),
                    sign * (radius - Ri * cos(ang)))
                ]

                section = list(range(0, len(p1), Nmax))
                if not polyline:
                    for s in section[:-1]:
                        outline = pstart\
                            + p1[s:s + Nmax]\
                            + list(reversed(p2[s:s + Nmax]))
                        nd.Polygon(layer=lay, points=outline).put(0)
                        pstart = [p2[s + Nmax - 1], p1[s + Nmax - 1]]

                    outline = pstart\
                         + p1[section[-1]:section[-1] + Nmax]\
                         + pend\
                         + list(reversed(p2[section[-1]:section[-1] + Nmax]))
                    nd.Polygon(layer=lay, points=outline).put(0)
                else:
                    for s in section[:-1]:
                        centreline = [pstart[0]] + p1[s:s + Nmax]
                        nd.Polyline(layer=lay, points=centreline, width=wpoly).put(0)
                        pstart = [p2[s + Nmax - 1], p1[s + Nmax - 1]]

                    centreline = [pstart[0]]\
                        + p1[section[-1]:section[-1] + Nmax]\
                        + [pend[0]]
                    nd.Polyline(layer=lay, points=centreline, width=wpoly).put(0)


        C.properties['parameters'] = {
            'angle':  {'value': angle ,'unit': 'deg', 'type': 'float'},
            'width':  {'value': width, 'unit': 'um', 'type': 'float'},
            'radius': {'value': radius, 'unit': 'um', 'type': 'float'},
            'xs':     {'value': xs, 'unit': None, 'type': 'str'},
            'call':   {'value': 'bend', 'unit': None, 'type': 'str'},
        }
        return C
    return cell


def Tp_sinecurve(width=1, distance=200, offset=20, xs=None, layer=None,
        name=None):
    """Template for creating parametrized sine curve waveguide function.

    Args:
        width (float): width of the interconnect in um
        pin (Node): optional Node for modeling info
        xs (str): xsection of sinebend
        distance (float): total forward length of the sinebend in um
        offset (float): lateral offset of the sinebend in um
        xs (str): xsection of waveguide
        layer (int | str): layer number or layer name

    Returns:
        function: Function returning a Cell object with the sinecurve guide
    """
    # TODO: add growy
    def cell(width=width, distance=distance, offset=offset, xs=xs,
            layer=layer, name=name):
        """Create a (raised) sine bend waveguide element.

        Args:
            width (float): width of the interconnect in um
            pin (Node): optional Node for modeling info
            xs (str): xsection of sinebend
            offset (float): lateral offset of the sinebend in um
            distance (float): total forward length of the sinebend in um
            xs (str): xsection of waveguide
            layer (int | str): layer number or layer name

        Returns:
            Cell: sinecurve element
        """

        if name is None:
            name = 'sinecurve'

        xya = (distance, offset, 0) # End point
        # sinecurve waveguide, cwg
        with Cell(name=name, cnt=True) as C:
            C.instantiate = cfg.instantiate_mask_element
            nd.Pin(name='a0', type=0, width=width, xs=xs, show=True).put(0, 0, 180)
            nd.Pin(name='b0', type=1, width=width, xs=xs, show=True).put(*xya)
            # TODO: length_geo
            for lay, grow, acc, polyline in layeriter(xs, layer):
                (a1, b1), (a2, b2), c1, c2 = grow
                (a1, b1), (a2, b2) = (-a1, -b1), (-a2, -b2)
                # TODO: define polyline2polygon with left and right width from spine
                # sampled curve
                xy = curve2polyline(sinebend_point, xya, acc, (distance, offset))
                # polygon of proper width
                xy = nd.util.polyline2polygon_asym(xy, width, grow=((a1, b1), (a2, b2)))
                nd.Polygon(layer=lay, points=xy).put(0)
        return C
        C.properties['parameters'] = {
            'distance':  {'value': distance ,'unit': 'um', 'type': 'float'},
            'width':  {'value': width, 'unit': 'um', 'type': 'float'},
            'offset': {'value': offset, 'unit': 'um', 'type': 'float'},
            'xs':     {'value': xs, 'unit': None, 'type': 'str'},
            'call':   {'value': 'sinecurve', 'unit': None, 'type': 'str'},
        }
    return cell


def Tp_cobra(xya=(100, 100, 10), width1=1.0, width2=1.0, radius1=0, radius2=0,
        offset1=None, offset2=None, xs=None, layer=None, name=None):
    """Template for creating parametrized cobra waveguide function.

    Args:
        xya (point): point to connect to from (0,0,0)
        width1 (float): width of waveguide
        width2 (float): width of waveguide at end
        radius1 (float): radius at start (0 is no curvature)
        radius2 (float): radius at end (0 is no curvature)
        offset1 (float | function): positive offset reduces radius.
            The offset can be a function F(width, radius) that returns a float
        offset2 (float | function): positive offset reduces radius.
            The offset can be a function F(width, radius) that returns a float
        xs (str): xsection of waveguide
        layer (int | str): layer number or layer name

    Returns:
        function: Function returning a Cell object with the cobra guide
    """
    # TODO: add growy
    def cell(xya=xya, width1=width1, width2=width2, radius1=radius1,
            radius2=radius2, offset1=offset1, offset2=offset2, xs=xs,
            layer=layer, name=name):
        """Create a parametric waveguide element.

        Args:
            xya (point): point to connect to from (0,0,0)
            width (float): width of waveguide
            xs (str): xsection of waveguide
            layer (int | str): layer number or layer name

        Returns:
            Cell: cobra element
        """
        if name is None:
            name = 'cobra'
        if offset1 is None:
            offset1 = __get_offset(xs, width1, radius1)
        if radius1 != 0:
            radius1 = radius1 - offset1
        if offset2 is None:
            offset2 = __get_offset(xs, width2, radius2)
        if radius2 != 0:
            radius2 = radius2 - offset2
        if width2 is None:
            width2 = width1

        # cobra waveguide, pwg
        with Cell(name=name, cnt=True) as C:
            C.instantiate = cfg.instantiate_mask_element
            p1 = nd.Pin(name='a0', type=0, width=width1, xs=xs, show=True).\
                put(0, -offset1, 180)
            xya = (xya[0], xya[1]-offset1, xya[2])
            p2 = nd.Pin(name='b0', type=1, width=width2, xs=xs, show=True).put(*xya)

            xya = (xya[0] - offset2*sin(np.radians(xya[2])), xya[1] +
                offset2*cos(np.radians(xya[2])), xya[2])
            # Solve the generic bend
            A, B, L, Rmin = gb_coefficients(xya, radius1=radius1,
                radius2=radius2)
            nd.connect_path(p1, p2, None)
            C.length_geo = None
            C.properties['Rmin'] = Rmin
            saved_acc = 10

            Rdrc = 1e8
            try:
                Rdrc = cfg.XSdict[xs].minimum_radius
                if Rdrc is None:
                    Rdrc = 0
            except KeyError:
                if xs not in cfg.XSdict.keys():
                    Rdrc = 0
            if Rdrc > Rmin:
                logger.warning('DRC minimum_radius {:.3f} < {:.3f}'.
                    format(Rmin, Rdrc))

            for lay, grow, acc, polyline in layeriter(xs, layer):
                (a1, b1), (a2, b2), c1, c2 = grow
                (a1, b1), (a2, b2) = (-a1, -b1), (-a2, -b2)
                # sampled curve
                xy = curve2polyline(gb_point, xya, acc, (A, B, L))
                if C.length_geo is None or saved_acc > acc:
                    # Calculate length of polyline and keep the one for the
                    # most accurate layer
                    C.length_geo = polyline_length(xy)
                    saved_acc = acc
                # polygon of proper width
                # TODO: add growy
                if callable(width1):
                    def wfun(t):
                        return width1(t)
                    xy2 = nd.util.polyline2polygon_asym(
                        xy,
                        wfun,
                        grow=((a1, b1), (a2, b2))
                    )
                else:
                    xy2 = nd.util.polyline2polygon_asym(
                        xy,
                        width1,
                        width2,
                        grow=((a1, b1), (a2, b2))
                    )
                nd.Polygon(layer=lay, points=xy2).put(0)
        return C
    return cell


#TODO: not supposed to be defined in this template module.
    # See also show_pin()
wherecnt = count()
def whereami(text='here', size=100, pin=None):
    """Show current pointer position as arrow in the layout.

    Args:
        text (str): annotation text
        size (float): size of the annotation

    Returns:
        None
    """
    layer = 500
    nd.cp.push()
    if pin is None:
        pin = nd.cp.here()
    points = [
        (0, 0), (-0.5, 0.5), (-0.4, 0.25), (-1, 0.3), (-1, -0.3),
        (-0.4, -0.25), (-0.5, -0.5)]
    points = [(x*size,y*size) for x, y in points]
    with Cell('I am'.format(wherecnt)) as crisis:
        nd.Polygon(layer=layer, points=points).put(0)
        nd.text(text+' ', height=size/4, align='rc', layer=layer).put(0)

    crisis.put(pin)
    nd.cp.pop()


# ==============================================================================
# create elements
# ==============================================================================
strt = Tp_straight()
bend = Tp_arc()
ptaper = Tp_ptaper()
taper = Tp_taper()
cobra = Tp_cobra()
sinebend = Tp_sinecurve()


# ==============================================================================
# #create a default cell and set cp.
# ==============================================================================
cfg.defaultcell = Cell(name=cfg.defaultcellname)
cfg.cp = cfg.defaultcell.org
